"""The :py:mod:`ssb_timeseries.sample_data` module provides tools for creating sample timeseries data. These are convenience functions for tests and demos."""

import itertools
from datetime import datetime

import numpy as np
import pandas as pd

from ssb_timeseries.dates import date_utc

# mypy: disable-error-code="arg-type, type-arg, import-untyped, unreachable, attr-defined"


def series_names(*args: dict | str | list[str] | tuple, **kwargs: str) -> list[str]:
    """Return all permutations of the elements of multiple groups of strings.

    Args:
        *args (str | list | tuple | dict): Each arg in args should be a collection of names to be combined with the other.
        **kwargs (str): One option: 'separator' defines a character sequence inserted between name elements. Defaults to '_'.

    Returns:
        list[str]: List of names to be used as series names.

    Raises:
        ValueError: If an argument of an invalid type is passed.

    """
    separator = kwargs.get("separator", "_")

    if isinstance(args, dict):
        # TODO: fix mypy error: Statement is unreachable  [unreachable]
        return [value for value in args.values()]

    final_args = []

    for arg in args:
        if arg is None:
            final_args.append([""])
        if isinstance(arg, str):
            final_args.append([arg])
        elif isinstance(arg, list):
            final_args.append(arg)
        elif isinstance(arg, tuple):
            final_args.append(list(arg))
        else:
            raise ValueError(f"Invalid argument type: {type(arg)}")

    names = [
        separator.join(combination) for combination in itertools.product(*final_args)
    ]

    return names


def create_df(
    *lists: dict | list[str] | tuple | str,
    start_date: datetime | str | None = None,
    end_date: datetime | str | None = None,
    freq: str = "D",
    interval: int = 1,
    separator: str = "_",
    midpoint: int | float = 100,
    variance: int | float = 10,
    temporality: str = "AT",
    decimals: int = 0,
) -> pd.DataFrame:
    """Generate sample data for specified date range and permutations over lists.

    Args:
        start_date (datetime): The start date of the date range. Optional, default is negative infinity.
        end_date (datetime): The end date of the date range. Optional, default is positive infinity.
        *lists (list[str]): Lists of values to generate combinations from.
        freq (str): The frequency of date generation.
            'Y' for yearly at last day of year,
            'YS' for yearly at first day of year,
            'M' for monthly at last day of month,
            'MS' for monthly at first day of month,
            'W' for weekly on Sundays,
            'D' for daily,
            'H' for hourly,
            'T' for minutely,
            'S' for secondly,
            etc.
            Optional, default is 'D'.
        interval (int): The interval between dates. ; optionalDefault is 1.
        separator (str): The separator used to join combinations. Optional, default is '_'.
        midpoint (float): The midpoint value for generating random data. Optional, default is 100.
        variance (float): The variance value for generating random data. Optional, default is 10.
        temporality (str): The temporality of the data. Default is 'AT'.
        decimals (int): The number of decimal places to round to. Optional, default is 0.

    Returns:
        DataFrame: A DataFrame containing sample data.

    Example:
    ```
    # Generate sample data with no specified start or end date (defaults to +/- infinity)
    sample_data = generate_sample_df(List1, List2, freq='D')
    ```
    """
    # Handle start_date and end_date defaults
    if start_date is None:
        start_date = datetime.min  # Representing negative infinity
    if end_date is None:
        end_date = datetime.max  # Representing positive infinity

    # Add other frequencies as needed
    freq_lookup = {
        "Y": "years",
        "YS": "years",
        "M": "months",
        "ME": "months",
        "MS": "months",
        "D": "days",
        "H": "hours",
        "T": "minutes",
        "S": "seconds",
    }
    offset = pd.DateOffset(**{freq_lookup[freq]: interval})
    valid_at = pd.date_range(start=start_date, end=end_date, freq=f"{interval}{freq}")
    valid_from = valid_at
    valid_to = pd.date_range(
        start=date_utc(start_date) + offset,
        end=date_utc(end_date) + offset,
        freq=f"{interval}{freq}",
    )
    # ... valid_to = valid_from + pd.DateOffset(**{freq_lookup[freq]: interval})  # type: ignore

    # BUGFIX: If *lists receives strings, permutations will be over chars by chars
    # Kombiner listene til en enkelt liste av lister
    # list = list(lists)

    # name_parts = series_names(*lists)
    # Generer alle mulige kombinasjoner av listene med separator
    # series = [
    #     separator.join(combination) for combination in itertools.product(*name_parts)
    # ]
    series = series_names(*lists, separator=separator)

    # Opprett DataFrame med tilfeldige tall
    rows = len(valid_at)
    cols = len(series)
    some_numbers = random_numbers(
        rows, cols, midpoint=midpoint, variance=variance, decimals=decimals
    )
    df = pd.DataFrame(
        # (midpoint + variance * np.random.randn(rows, cols)).round(decimals),
        some_numbers,
        columns=series,
        dtype="float32[pyarrow]",
    )

    # Legg til "Dates" som den første kolonnen i "df"
    match temporality:
        case "AT":
            df.insert(0, "valid_at", valid_at)
            df.set_index("valid_at")
        case "FROM_TO":
            df.insert(0, "valid_to", valid_to)
            df.insert(0, "valid_from", valid_from)
            df.set_index(["valid_from", "valid_to"])

    return df


def random_numbers(
    rows: int,
    cols: int,
    decimals: int = 0,
    midpoint: int | float = 100,
    variance: int | float = 10,
) -> np.ndarray:
    """Generate sample dataframe of specified dimensions."""
    generator = np.random.default_rng()
    random_matrix = generator.standard_normal(size=(rows, cols))
    return midpoint + variance * random_matrix.round(decimals)



import os

# from ssb_timeseries import config

"""Use this module for running / validating setups?
ISSUE: the code is not found when trying to run `python timeseries`
... but is working fine for `python -m timeseries`
--> have a closer look at pyproject.toml?
"""


def main() -> None:
    """Validate setup."""
    config_file = os.environ["TIMESERIES_CONFIG"]
    greeting = f"Hello Timeseries!\n... the configuration file is: {config_file}. \nAdditional set up steps may be added later!"

    print(greeting)
    # perform set up steps:
    # os.environ["TIMESERIES_CONFIG"] = DEFAULT_CONFIG_LOCATION


if __name__ == "__main__":
    """Running `python timeseries` or `python -m timeseries` should run or validate setup."""
    main()



"""Time series should make use of standardised messaging and logging.

A key principle is that every dataset read or write operation should be logged with log level INFO. The plan is to take this one step further and add stack inspection to enrich the log messages with process identifiers. In that way, logs would provide lineage at the dataset level. Also, adding a log sink to put the (dataset write) messages on a message queue would allow event based workflow definitions.

These are features that may warrant their own shared Dapla libraries. This module serves as a specification of required features byd separates out code that could go into such a library.
"""

import functools
import logging
from datetime import datetime

from typing_extensions import Self

from ssb_timeseries.config import Config
from ssb_timeseries.types import F

# nosonar: disable comment
# import uuid
# automatic cloud logging config
# import google.cloud.logging
# client = google.cloud.logging.Client()
# client.setup_logging()

# mypy: disable-error-code="operator, no-untyped-def, return-value, import-untyped, arg-type"
# ruff: noqa: ANN002, ANN003

ts_logger = logging.getLogger("TIMESERIES")
log_string = logging.Formatter(
    "%(name)s | %(levelname)s | %(asctime)s | %(message)s \n"
)
log_json = logging.Formatter(
    '{"name": "%(name)s"; "level": %(levelname)s; "timestamp": %(asctime)s; "message": "%(message)s" }'
)

# nosonar: disable comment
# CONFIGURATION_FILE = str(os.environ.get("TIMESERIES_CONFIG"))
# CONFIG = Config(configuration_file=CONFIGURATION_FILE)
LOG_FILE = str(Config.active().log_file)


file_handler = logging.FileHandler(LOG_FILE)
file_handler.setFormatter(log_json)
file_handler.setLevel(logging.INFO)
ts_logger.addHandler(file_handler)

# Also log to console.
console = logging.StreamHandler()
console.setFormatter(log_string)  # BUG: format does not take effect in console?
console.setLevel(logging.WARNING)
ts_logger.addHandler(console)

# nosonar: disable comment
# Google Cloud logging:
# from google.cloud.logging.handlers import CloudLoggingHandler
# cloud_handler = CloudLoggingHandler(client)
# ts_logger.addHandler(cloud_handler)


class EnterExitLog:
    """Class supporting decorator to log on enter and exit."""

    def __init__(self, name: str) -> None:
        """Enter/exit template for workflow process."""
        self.name = name

    def __enter__(self) -> Self:
        """Before each workflow process step, do this."""
        self.init_time = datetime.now()
        ts_logger.info(f"START: {self.name}.")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:  # noqa ANN001
        """After each workflow process step, do this."""
        self.end_time = datetime.now()
        self.elapsed_time = self.end_time - self.init_time
        ts_logger.info(
            f"FINISH: {self.name}. Completed in: {self.elapsed_time} seconds."
        )


def log_start_stop(func: F) -> F:
    """Log start and stop of decorated function."""
    # nosonar  TODO: generalise: pass in functions to enter/exit?

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with EnterExitLog(func.__name__):
            out = func(*args, **kwargs)

        return out

    return wrapper


# nosonar disable comment
# def debug(func):  # ANN001, ANN201
#     """Print the function signature and return value."""

#     @functools.wraps(func)
#     def wrapper_debug(*args, **kwargs):
#         args_repr = [repr(a) for a in args]  # 1
#         kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]  # 2
#         signature = ", ".join(args_repr + kwargs_repr)  # 3
#         print(f"Calling {func.__name__}({signature})")
#         value = func(*args, **kwargs)
#         print(f"{func.__name__!r} returned {value!r}")  # 4
#         return value
#     return wrapper_debug
# """
#
# # @wraps??
# class Timer:
#     def __init__(self, name: str):
#         self.name: str = name
#
#     def __enter__(self):
#         self.init_time = datetime.now()
#         ts_logger.info("Started: {self.name}.")
#         return self
#
#     def __exit__(self, type, value, tb):
#         self.end_time = datetime.now()
#         self.elapsed_time = self.end_time - self.init_time
#         ts_logger.info("Finished: {self.name} in: {self.elapsed_time} seconds.")
#
# def funcion_timer(func) -> Callable[..., Function]:
#     @functools.wraps(func)
#     def wrapper(*args, **kwargs) -> Function:
#         with Timer(func.__name__):
#             return func(*args, **kwargs)
#
#     return wrapper
# """



"""The :py:mod:`ssb_timeseries.dataset` module and its :py:class:`Dataset` class is the very core of the :py:mod:`ssb_timeseries` package, defining most of the key functionality.

The dataset is the unit of analysis for both :doc:`information model <../info-model>` and :doc:`workflow integration <../workflow>`,and performance will benefit from linear algebra with sets as matrices consisting of series column vectors.

As described in the :doc:`../info-model` time series datasets may consist of any number of series of the same :py:class:`type <ssb_timeseries.properties.SeriesType>`. Series types are defined by :py:class:`properties.Versioning` and :py:class:`properties.Temporality`, see :py:class:`properties.SeriesType`.

It is also strongly encouraged to make sure that the resolutions of the series in datasets are the same, and to minimize the number of gaps in the series. Very sparse data is a strong indication that a dataset is not well defined: may indicate that series in the set have different origins. What counts as 'gaps' in this context is any representation of undefined values: None, null, NAN or "not a number" values, as opposed to the number zero. The number zero is a gray area - it can be perfectly valid, but can also be an indication that not all the series should be part of the same set.

.. seealso::
    See documentation for the :py:mod:`ssb_timeseries.catalog` module for tools for searching for datasets or series by names or metadata.

"""

import re
from copy import deepcopy
from datetime import datetime
from typing import Any
from typing import Protocol
from typing import no_type_check

import matplotlib.pyplot as plt  # noqa: F401
import numpy as np
import pandas as pd
from typing_extensions import Self

from ssb_timeseries import io
from ssb_timeseries import meta
from ssb_timeseries import properties
from ssb_timeseries.dates import date_local
from ssb_timeseries.dates import date_utc  # type: ignore[attr-defined]
from ssb_timeseries.dates import utc_iso  # type: ignore[attr-defined]
from ssb_timeseries.logging import ts_logger
from ssb_timeseries.types import F
from ssb_timeseries.types import PathStr

# mypy: disable-error-code="assignment,attr-defined,union-attr,arg-type,call-overload,no-untyped-call,dict-item"
# ruff: noqa: RUF013


class IO(Protocol):
    """Interface for IO operations."""

    def save(self) -> None:
        """Save the dataset."""
        ...

    def snapshot(self) -> None:
        """Save a snapshot of the dataset."""
        ...


class Dataset:
    """Datasets are the core unit of analysis for workflow and data storage.

    A dataset is a logical collection of data and metadata stemming from the same process origin. All the series in a dataset must be of the same type.

    The type defines
         * versioning (NONE, AS_OF, NAMED)
         * temporality (Valid AT point in time, or FROM and TO for duration)
         * value (for now only scalars)

    .. seealso:
        :doc:`../info-model`

    """

    def __init__(
        self,
        name: str,
        data_type: properties.SeriesType = None,
        as_of_tz: datetime = None,
        load_data: bool = True,
        **kwargs: Any,
    ) -> None:
        """**Initialising** a dataset object can either retrieve data and metadata for an existing set or prepare a new one.

        If data_type versioning is specified as AS_OF, a datetime with timezone should be provided.
        If not, but data is passed, ::py:meth:`as_of_tz` defaults to current time. Providing an AS_OF date has no effect if versioning is NONE.

        When loading existing sets, load_data = false can be set in order to suppress reading large amounts of data.
        For data_types with AS_OF versioning, not providing the AS_OF date will have the same effect.

        Metadata will always be read.
        Data is represented as Pandas dataframes; but Polars lazyframes or Pyarrow tables is likely to be better.
        Initial implementation assumes stores data in parquet files, but feather files and various database options are considered for later.

        Support for additional "type" features/flags behaviours like sparse data may be added later (if needed).

        Data is kept in memory and not stored before explicit call to .save.
        """
        self.name: str = name
        if data_type:
            self.data_type = data_type
        else:
            look_for_it = search(name)
            if isinstance(look_for_it, Dataset):
                self.data_type = look_for_it.data_type
            else:
                raise ValueError(
                    f"Dataset {name} did not give a unique match. Specify data_type if you intend to initialise a new set."
                )
        identify_latest: bool = (
            self.data_type.versioning == properties.Versioning.AS_OF
            and not as_of_tz
            and load_data
        )
        if identify_latest:
            ts_logger.debug(
                f"Init {self.data_type} '{self.name}' without 'as_of_tz' --> identifying latest."
            )
            self.io = io.FileSystem(
                set_name=self.name, set_type=self.data_type, as_of_utc=None
            )
            lookup_as_of = self.versions()[-1]
            if isinstance(lookup_as_of, datetime):
                as_of_tz = lookup_as_of
            self.as_of_utc = date_utc(as_of_tz)
        else:
            self.as_of_utc = date_utc(as_of_tz)

        self.io: IO = io.FileSystem(  # type: ignore[no-redef]
            set_name=self.name,
            set_type=self.data_type,
            as_of_utc=self.as_of_utc,
        )

        # If data is provided by kwarg, use it.
        # Otherwise, load it ... unless explicitly told not to.
        kwarg_data: pd.DataFrame = kwargs.get("data", pd.DataFrame())
        if not kwarg_data.empty:
            self.data = kwarg_data
        elif load_data:  # and self.data_type.versioning == properties.Versioning.AS_OF:
            self.data = self.io.read_data(self.as_of_utc)
        else:
            self.data = pd.DataFrame()

        self.tags = self.default_tags()
        self.tags.update(self.io.read_metadata())
        self.tag_dataset(tags=kwargs.get("dataset_tags", {}))

        # all of the following should be turned into set level tags - or find another way into the parquet files?

        # autotag:
        self.auto_tag_config = {
            "name_pattern": kwargs.get("name_pattern", ""),
            "separator": kwargs.get("separator", "_"),
            "regex": kwargs.get("regex", ""),
            "apply_to_all": kwargs.get("series_tags", {}),
        }
        if not self.data.empty:
            self.tag_series(tags=self.auto_tag_config["apply_to_all"])
            self.series_names_to_tags()

        # "owner" / sharing / access
        self.product: str = kwargs.get("product", "")
        self.process_stage: str = kwargs.get("process_stage", "")
        self.sharing: dict[str, str] = kwargs.get("sharing", {})

    def copy(self, new_name: str, **kwargs: Any) -> Self:
        """Create a copy of the Dataset.

        The copy need to get a new name, but unless other information is spcecified, it will be create wiht the same data_type, as_of_tz, data, and tags.
        """
        out = deepcopy(self)
        for k, v in kwargs.items():
            setattr(out, k, v)

        out.rename(new_name)
        return out

    def rename(self, new_name: str) -> None:
        """Rename the Dataset.

        For use by .copy, and on very rare other occasions. Does not move or rename any previously stored data.
        """
        self.name = new_name

        self.tags["name"] = new_name
        for _, v in self.tags["series"].items():
            v["dataset"] = new_name  # type: ignore

    def save(self, as_of_tz: datetime = None) -> None:
        """Persist the Dataset.

        Args:
            as_of_tz (datetime): Provide a timezone sensitive as_of date in order to create another version. The default is None, which will save with Dataset.as_of._utc (utc dates under the hood).
        """
        if as_of_tz is not None:
            self.as_of_utc = date_utc(as_of_tz)

        self.io = io.FileSystem(self.name, self.data_type, self.as_of_utc)
        # ts_logger.debug(f"DATASET {self.name}: SAVE. Tags:\n\t{self.tags}.")
        if not self.tags:
            self.tags = self.default_tags()
            self.tags.update(self.io.read_metadata())

            # autotag:
            if not self.data.empty:
                self.series_names_to_tags()
                ts_logger.debug(
                    f"DATASET {self.name}: attempt to save empty tags = {self.tags}."
                )

        self.io.save(meta=self.tags, data=self.data)

    def snapshot(self, as_of_tz: datetime = None) -> None:
        """Copy data snapshot to immutable processing stage bucket and shared buckets.

        Args:
            as_of_tz (datetime): Optional. Provide a timezone sensitive as_of date in order to create another version. The default is None, which will save with Dataset.as_of_utc (utc dates under the hood).
        """
        date_from = self.data[self.datetime_columns()].min().min()
        date_to = self.data[self.datetime_columns()].max().max()
        ts_logger.debug(
            f"DATASET {self.name}: Data {utc_iso(date_from)} - {utc_iso(date_to)}:\n{self.data.head()}\n...\n{self.data.tail()}"
        )

        self.save(as_of_tz=self.as_of_utc)

        self.io.snapshot(
            product=self.product,
            process_stage=self.process_stage,
            sharing=self.sharing,
            period_from=date_from,
            period_to=date_to,
        )

    def versions(self, **kwargs: Any) -> list[datetime | str]:
        """Get list of all series version markers (`as_of` dates or version names).

        By default `as_of` dates will be returned in local timezone. Provide `return_type = 'utc'` to return in UTC, 'raw' to return as-is.
        """
        versions = self.io.list_versions(
            file_pattern="*.parquet", pattern=self.data_type.versioning
        )
        if not versions:
            return []
        else:
            ts_logger.debug(f"DATASET {self.name}: versions: {versions}.")

        if self.data_type.versioning == properties.Versioning.AS_OF:
            return_type = kwargs.get("return_type", "local")
        else:
            return_type = "raw"

        match return_type:
            case "local":
                return [date_local(v) for v in versions]
            case "utc":
                return [date_utc(v) for v in versions]
            case "utc_iso":
                return [utc_iso(v) for v in versions]
            case _:
                return versions

    @property
    def series(self) -> list[str]:
        """Get series names."""
        if (
            self.__getattribute__("data") is None
            and self.__getattribute__("tags") is None
        ):
            return sorted(self.series_tags.keys())
        else:
            return sorted(self.numeric_columns())

    @property
    def series_tags(self) -> meta.SeriesTagDict:
        """Get series tags."""
        return self.tags["series"]  # type: ignore

    def default_tags(self) -> meta.DatasetTagDict:
        """Return default tags for set and series."""
        return {
            "name": self.name,
            "versioning": str(self.data_type.versioning),
            "temporality": str(self.data_type.temporality),
            "series": {s: {"dataset": self.name, "name:": s} for s in self.series},
        }

    def tag_dataset(
        self,
        tags: meta.TagDict = None,
        **kwargs: str | list[str] | set[str],
    ) -> None:
        """Tag the set.

        Tags may be provided as dictionary of tags, or as kwargs.

        In both cases they take the form of attribute-value pairs.

        Attribute (str): Attribute identifier.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        Examples:
            **Dependencies**

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.properties import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>>
            >>> x = Dataset(name='sample_dataset',
            >>>         data_type=SeriesType.simple(),
            >>>         data=create_df(['x','y','z'],
            >>>             start_date='2024-01-01',
            >>>             end_date='2024-12-31',
            >>>             freq='MS',)
            >>> )
            >>>
            >>> x.tag_dataset(tags={'country': 'Norway', 'about': 'something_important'})
            >>> x.tag_dataset(another_attribute='another_value')

            Note that while no such restrictions are enforced, it is strongly recommended that both attribute names (``keys``) and ``values`` are standardised.
            The best way to ensure that is to use taxonomies (for SSB: KLASS code lists). However, custom controlled vocabularies can also be maintained in files.i

            .. seealso::
                :doc:`../info-model`
                :py:class:`Dataset.tag_series`
        """
        if not self.__getattribute__("tags"):
            # should not be possible, hence
            raise ValueError(f"Tags not defined for dataset: {self.name}.")

        if not tags and not kwargs:
            return
        elif not tags:
            tags = {}
        if kwargs:
            tags.update(**kwargs)

        propagate = tags.pop("propagate", True)
        if tags:
            self.tags = meta.add_tag_values(self.tags, tags, recursive=propagate)

    def tag_series(
        self,
        identifiers: str | list[str] | None = None,
        name_pattern: list[str] | None = None,
        separator: str = "_",
        tags: meta.TagDict = None,
        **kwargs: str | list[str],
    ) -> None:
        """Tag the series identified by ``identifiers`` with provided tags.

        Tags may be provided as dictionary of tags, or as kwargs.

        In both cases they take the form of attribute-value pairs.

        Attribute (str): Attribute identifier.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        If series names follow the same pattern of attribute values in the same order separated by the same character sequence, tags can be propagated accordingly by specifying ``name_pattern`` and ``separator`` parameters. The separator will default to underscore if not provided. Note that propagation by pattern will affect *all* series in the set, not only the ones identified by ``identifiers``.

        Examples:
            **Dependencies**

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.properties import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>>
            >>> some_data = create_df(['x', 'y', 'z'], start_date='2024-01-01', end_date='2024-12-31', freq='MS')

            **Tag by kwargs**

            >>> x = Dataset(name='sample_set',data_type=SeriesType.simple(),data=some_data)
            >>> x.tag_series(example_1='string_1', example_2=['a', 'b', 'c'])

            **Tag by dict**

            >>> x = Dataset(name='sample_set',data_type=SeriesType.simple(),data=some_data)
            >>> x.tag_series(tags={'example_1': 'string_1', 'example_2': ['a', 'b', 'c']})
        """
        if not tags:
            tags = {}
        tags.update(kwargs)

        if not identifiers:
            identifiers = self.series

        inherit_from_set_tags = meta.inherit_set_tags(self.tags)

        for ident in identifiers:
            self.tags["series"][ident] = {
                "name": ident,
            }
            self.tags["series"][ident].update({**inherit_from_set_tags, **tags})

        # if name_pattern:
        #    self.series_names_to_tags(attributes=name_pattern, separator=separator)

    def detag_dataset(
        self,
        *args: str,
        **kwargs: Any,
    ) -> None:
        """Detag selected attributes of the set.

        Tags to be removed may be provided as list of attribute names or as kwargs with attribute-value pairs.
        """
        self.tags = meta.delete_dataset_tags(
            self.tags,
            *args,  # [a for a in args if isinstance(a, str)]
            propagate=True,
            **kwargs,
        )

    @no_type_check  # "operator"
    def detag_series(
        self,
        *args: str,
        **kwargs: Any,
    ) -> None:
        """Detag selected attributes of series in the set.

        Tags to be removed may be specified by args or kwargs.
        Attributes listed in `args` will be removed from all series.

        For kwargs, attributes will be removed from the series if the value matches exactly. If the value is a list, the matching value is removed. If kwargs contain all=True, all attributes except defaults are removed.
        """
        self.tags["series"] = meta.delete_series_tags(
            self.tags["series"],
            *args,
            **kwargs,
        )

    @no_type_check
    def series_names_to_tags(
        self,
        attributes: list[str] | None = None,  # /NOSONAR
        separator: str = "",  # /NOSONAR
        regex: str = "",  # /NOSONAR
    ) -> None:
        """Tag all series in the dataset based on a series 'attributes', ie a list of attributes matching positions in the series names when split on 'separator'.

        Alternatively, a regular expression with groups that match the attributes may be provided.
        Ideally attributes relies on KLASS, ie a KLASS taxonomy defines the possible attribute values.

        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.
        Value (str): Element identifier, unique within the taxonomy. Ideally KLASS code.

        Example:
            **Dependencies**

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.properties import SeriesType
            >>> from ssb_timeseries.sample_data import create_df

            **Tag using name_pattern**

            If all series names follow a uniform pattern where attribute values are separated by the same character sequence:

            >>> some_data = create_df(["x_a", "y_b", "z_c"], start_date="2024-01-01", end_date="2024-12-31", freq="MS",)
            >>> x = Dataset(name="sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=some_data)
            >>> x.series_names_to_tags(attributes=['XYZ', 'ABC'])

            **Tag by regex**

            If series names are less well formed, a regular expression with groups matching the attribute list can be provided instead of the separator parameter.

            >>> more_data = create_df(["x_1,,a", "y...b..", "z..1.1-23..c"], start_date="2024-01-01", end_date="2024-12-31", freq="MS")
            >>> x = Dataset(name="sample_set",data_type=SeriesType.simple(),data=more_data,)
            >>> x.series_names_to_tags(attributes=['XYZ', 'ABC'], regex=r'([a-z])*([a-z])')


            The above approach may be used to add tags for an existing dataset, but the same arguments can also be provided when initialising the set:

            >>> # xdoctest: +SKIP
            >>> # this line is not valid python code
            >>> z = Dataset(name="sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=some_data,
            >>>     name_pattern=['XYZ', 'ABC'])
            >>> # xdoctest: -SKIP

            Best practice is to do this only in the process that writes data to the set. For a finite number of series, it does not need to be repeated.

            If, on the other hand, the number of series can change over time, doing so at the time of writing ensures all series are tagged.
        """
        # if attributes is None:
        #     attributes = []

        # if attributes:
        #     self.auto_tag_config["name_pattern"] = attributes
        # else:
        #    attributes = self.auto_tag_config["name_pattern"]
        apply_all = self.auto_tag_config.get("apply_to_all", {})
        inherited = meta.inherit_set_tags(self.tags)
        # self.tag_series({**inherited, **apply_all})

        # if separator:
        #     self.auto_tag_config["separator"] = separator
        # else:
        #    separator = self.auto_tag_config["separator"]

        for series_key in self.series:
            self.tags["series"].get(series_key, {}).update(inherited)
            if apply_all:
                self.tags["series"].get(series_key, {}).update(apply_all)

        attributes = self.auto_tag_config["name_pattern"]
        separator = self.auto_tag_config["separator"]
        regex = self.auto_tag_config["regex"]
        if attributes:
            for series_key in self.tags["series"].keys():
                if regex:
                    name_parts = re.search(regex, series_key)
                else:
                    name_parts = series_key.split(separator)

                for attribute, value in zip(attributes, name_parts, strict=False):
                    # not necessary? self.tags["series"][series_key][attribute] = deepcopy(value)
                    self.tags["series"][series_key][attribute] = value

    def replace_tags(
        self,
        *args: tuple[meta.TagDict, meta.TagDict],
    ) -> None:
        """Retag selected attributes of series in the set.

        The tags to be replaced and their replacements should be specified as tuple(s) of dictionaries for `(old_tags, new_tags)`. Both can contain multiple tags.
         * Each tuple is evaluated independently for each series in the set.
         * If the tag dict to be replaced contains multiple tags, all must match for tags to be replaced.
         * If the new tag dict contains multiple tags, all are added where there is a match.
        """
        for a in args:
            old = a[0]
            new = a[1]

            self.tags = meta.replace_dataset_tags(self.tags, old, new, recursive=True)

    @no_type_check
    def filter(
        self,
        pattern: str = "",
        tags: dict[str, str | list[str]] = None,
        regex: str = "",
        output: str = "dataset",
        new_name: str = "",
        **kwargs: str | list[str],
    ) -> pd.DataFrame | Self:
        """Filter dataset.data by textual pattern, regex or metadata tag dictionary.

        Or a combination.

        Args:
            pattern (str): Text pattern for search 'like' in column names. Defaults to ''.
            regex (str): Expression for regex search in column names. Defaults to ''.
            tags (dict): Dictionary with tags to search for. Defaults to None. All tags in dict must be satisfied for the same series (tags are combined by AND). If a list of values is provided for a tag, the criteria is satisfied for either of them (OR).
                | list(dict) Support for list(dict) is planned, not yet implemented, to satisfy alternative sets of criteria (the dicts will be combined by OR).
            output (str): Output type - dataset or dataframe.(df). Defaults to 'dataset'. Short forms 'df' or 'ds' are accepted.
            new_name (str): Name of new Dataset. If not provided, a new name is generated.
            **kwargs: if provided, goes into the init of the new set.

        Returns:
            Dataset | Dataframe:
            By default a new Dataset (a deep copy of self). If output="dataframe" or "df", a dataframe.
            TODO: Explore shallow copy / nocopy options.
        """
        if regex:
            df = self.data.filter(regex=regex).copy(deep=True)
            matching_series = df.columns

        if pattern:
            df = self.data.filter(like=pattern).copy(deep=True)
            matching_series = df.columns

        if tags:
            matching_series = meta.search_by_tags(self.tags["series"], tags)
            ts_logger.debug(f"DATASET.filter(tags) matched series:\n{matching_series} ")
            df = self.data[matching_series].copy(deep=True)

        df = pd.concat([self.data[self.datetime_columns()], df], axis=1)

        interval = kwargs.get("interval")
        if interval:
            ts_logger.warning(
                f"DATASET.__filter__: Attempted call with argument 'interval' = {interval} is not supported. --> TO DO!"
            )

        match output:
            case "dataframe" | "df":
                out = df
            case "dataset" | "ds" | _:
                if not new_name:
                    new_name = f"COPY of({self.name} FILTERED by pattern: {pattern}, regex: {regex} tags: {tags})"
                out = self.copy(new_name=new_name, data=df, **kwargs)
                matching_series_tags = {
                    k: v for k, v in out.tags["series"].items() if k in matching_series
                }
                out.tags["series"] = matching_series_tags
        return out

    @no_type_check
    def __getitem__(
        self, criteria: str | dict[str, str | list[str]] = "", **kwargs: Any
    ) -> Self | None:
        """Access Dataset.data.columns via Dataset[ list[column_names] | pattern | tags].

        Arguments:
            criteria:  Either a string pattern or a dict of tags.
            kwargs: If criteria is empty, this is passed to filter().

        Returns:
            Self | None

        Raises:
            TypeError: If filter() returns another type than Dataset.
        """
        # pattern: str = "", regex: str = "", tags: dict = {}):
        # Dataset[...] should return a Dataset object (?) with only the requested items (columns).
        # but should not mutate the original object, ie "self",
        # so that if x is a Dataset and x[a] a columnwise subset of x
        # x[a] *= 100 should update x[a] "inside" the original x, without "setting" x to z[a]
        # that later references to x should return the entire x, not only x[a].
        # Is this possible, or do we need to return a copy?
        # (Then the original x is not affected by updates to x[a])?
        # Or, is there a trick using dataframe views?
        # --->
        if criteria and isinstance(criteria, str):
            result = self.filter(pattern=criteria)
        elif criteria and isinstance(criteria, dict):
            result = self.filter(tags=criteria)
        elif kwargs:
            ts_logger.debug(f"DATASET.__getitem__(:\n\t{kwargs} ")
            result = self.filter(**kwargs)
        else:
            return None
        if isinstance(result, Dataset):
            return result
        else:
            raise TypeError("Dataset.filter() did not return a Dataset type.")

    def plot(self, *args: Any, **kwargs: Any) -> Any:
        """Plot dataset data.

        Convenience wrapper around Dataframe.plot() with sensible defaults.
        """
        df = self.data.copy()

        if self.data_type.temporality == properties.Temporality.FROM_TO:
            interval_handling = kwargs.pop("interval_handling", "interval").lower()
            match interval_handling:
                case "interval":
                    from_data = df
                    to_data = df
                    from_data["valid_to"] = from_data["valid_from"]
                    df = pd.concat(
                        [from_data, to_data],
                        axis=0,
                        ignore_index=True,
                    ).sort_values(by=["valid_from", "valid_to"])
                    df.drop(columns=["valid_to"], inplace=True)
                    xlabels = "valid_from"
                case "midpoint":
                    xlabels = "midpoint"
                    df["midpoint"] = df[self.datetime_columns()].median(axis=1)
                    df.drop(columns=["valid_from", "valid_to"], inplace=True)

                case _:
                    raise ValueError(
                        "Invalid option for interval_handling. Must be 'from', 'to', 'interval' or 'midpoint'."
                    )
        else:
            xlabels = self.datetime_columns()[0]

        ts_logger.debug(f"DATASET.plot(): x labels = {xlabels}")
        ts_logger.debug(f"Dataset.plot({args!r}, {kwargs!r}) x-labels {xlabels}")

        return df.plot(  # type: ignore[call-overload]
            xlabels,
            *args,
            legend=len(self.data.columns) < 9,
            title=self.name,
            figsize=(12, 4),
            **kwargs,
        )

    def vectors(self, pattern: str = "") -> None:
        """Get vectors with names equal to column names from Dataset.data.

        Args:
            pattern (str): Optional pattern for simple filtering of column names containing pattern. Defaults to ''.

        .. warning:: Caution!
            This (re)assigns variables in the scope of the calling function by way of stack inspection and hence risks of reassigning objects, functions, or variables if they happen to have the same name.

        """
        import inspect

        stack = inspect.stack()
        locals_ = stack[1][0].f_locals

        for col in self.data.columns:
            if col.__contains__(pattern):
                cmd = f"{col} = self.data['{col}']"
                ts_logger.debug(cmd)
                # the original idea was running (in caller scope)
                # exec(cmd)
                locals_[col] = self.data[col]

    def groupby(
        self,
        freq: str,
        func: str = "auto",
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Group dataset data by specified frequency and function.

        Returns a new Dataset.
        """
        datetime_columns = list(
            set(self.data.columns) & {"valid_at", "valid_to", "valid_from"}
        )
        ts_logger.debug(f"DATASET {self.name}: datetime columns: {datetime_columns}.")

        # works for datetime_columns = "valid_at", untested for others
        # TODO: add support for ["valid_from", "valid_to"]
        period_index = pd.PeriodIndex(self.data[datetime_columns[0]], freq=freq)
        ts_logger.debug(f"DATASET {self.name}: period index\n{period_index}.")

        # Fix for case when **kwargs contains numeric_only
        if "numeric_only" in kwargs:
            kwargs.pop("numeric_only")
        numeric_only_value = True

        match func:
            case "mean":
                out = self.data.groupby(period_index).mean(  # type: ignore[misc]
                    *args, numeric_only=numeric_only_value, **kwargs
                )
            case "sum":
                out = self.data.groupby(period_index).sum(  # type: ignore[misc]
                    *args, numeric_only=numeric_only_value, **kwargs
                )
            case "auto":
                # TODO: QA on exact logic / use "real" metadata
                # in particular, how to check meta data and blend d1 and df2 values as appropriate
                # (this implementation is just to show how it can be done)
                # QUESTION: do we need a default for "other" series / what should it be?
                df1 = self.data.groupby(period_index).mean(  # type: ignore[misc]
                    *args, numeric_only=numeric_only_value, **kwargs
                )
                ts_logger.debug(f"groupby\n{df1}.")

                df2 = (
                    self.data.groupby(period_index)
                    .sum(*args, numeric_only=numeric_only_value, **kwargs)  # type: ignore[misc]
                    .filter(regex="mendgde|volum|vekt")
                )
                ts_logger.debug(f"groupby\n{df2}.")

                df1[df2.columns] = df2[df2.columns]

                out = df1
                ts_logger.debug(f"groupby\n{out}.")
                ts_logger.debug(f"DATASET {self.name}: groupby\n{out}.")

        new_name = f"({self.name}.groupby({freq},{func})"

        return self.__class__(
            name=new_name,
            data_type=self.data_type,
            as_of_tz=self.as_of_utc,
            data=out,
        )

    def resample(
        self,
        freq: str,
        func: F | str,
        *args: Any,
        **kwargs: Any,
    ) -> Self:
        """Alter frequency of dataset data."""
        # TODO: have a closer look at dates returned for last period when upsampling
        # df = self.data.set_index(self.datetime_columns())
        df = self.data.set_index(self.datetime_columns()).copy()
        match func:
            case "min":
                out = df.resample(freq).min()
            case "max":
                out = df.resample(freq).max()
            case "sum":
                out = df.resample(freq).sum()
            case "mean":
                out = df.resample(freq).mean()
            case "ffill":
                out = df.resample(freq).ffill()
            case "bfill":
                out = df.resample(freq).bfill()
            case _:
                out = df.resample(freq, *args, **kwargs).apply(func)

        new_name = f"new set:[{self.name}.resampled({freq}, {func}]"
        return self.__class__(
            name=new_name,
            data_type=self.data_type,
            as_of_tz=self.as_of_utc,
            data=out,
        )

    # TODO: Add these?
    # def __iter__(self):
    #     self.n = 0 # start at latest version
    #     return self
    #
    # def __next__(self):
    #     if self.n <= self.data.columns:
    #         x = self.n
    #         self.n += 1 # return previous version
    #         return x
    #     else:
    #         raise StopIteration
    #     return x

    # TODO: rethink identity: is / is not behaviour
    # def identical(self, other:Self) -> bool:
    #     # check_data = self.__eq__(other:Self)
    #     # return all(check_defs) and check_data.all()
    #     return self.__dict__ == other.__dict__

    def all(self) -> bool:
        """Check if all values in series columns evaluate to true."""
        ts_logger.debug(all(self.data))
        return all(self.data[self.numeric_columns()])

    def any(self) -> bool:
        """Check if any values in series columns evaluate to true."""
        return any(self.data[self.numeric_columns()])

    def numeric_columns(self) -> list[str]:
        """Get names of all numeric series columns (ie columns that are not datetime)."""
        return list(set(self.data.columns).difference(self.datetime_columns()))

    def datetime_columns(self, *comparisons: Self | pd.DataFrame) -> list[str]:
        """Get names of datetime columns (valid_at, valid_from, valid_to).

        Args:
            *comparisons (Self | pd.DataFrame): Objects to compare with. If provided, returns the intersection of self and all comparisons.

        Returns:
            list[str]: The (common) datetime column names of self (and comparisons).

        Raises:
            ValueError: If comparisons are not of type Self or pd.DataFrame.
        """
        intersect = set(self.data.columns) & {"valid_at", "valid_from", "valid_to"}
        for c in comparisons:
            if isinstance(c, Dataset):
                intersect = set(c.data.columns) & intersect
            elif isinstance(c, pd.DataFrame):
                intersect = set(c.columns) & intersect
            else:
                raise ValueError(f"Unsupported type: {type(c)}")

        return list(intersect)

    @no_type_check
    def math(
        self,
        other: Self | pd.DataFrame | pd.Series | int | float,
        func,  # noqa: ANN001
    ) -> Any:
        """Generic helper making math functions work on numeric, non date columns of dataframe to dataframe, matrix to matrix, matrix to vector and matrix to scalar.

        Although the purpose was to limit "boilerplate" for core linear algebra functions, it also extend to other operations that follow the same differentiation pattern.

        Args:
            other (dataframe | series | matrix | vector | scalar ): One (or more?) pandas (polars to come) datframe or series, numpy matrix or vector or a scalar value.
            func (_type_): The function to be applied as `self.func(**other:Self)` or (in some cases) with infix notation `self f other`. Note that one or more date columns of the self / lefthand side argument are preserved, ie data shifting operations are not supported.

        Raises:
            ValueError: "Unsupported operand type"
            ValueError: "Incompatible shapes."

        Returns:
            Any:   Depending on the inputs: A new dataset / vector / scalar with the result. For datasets, the name of the new set is derived from inputs and the functions applied.
        """
        if isinstance(other, Dataset):
            ts_logger.debug(
                f"DATASET {self.name}: .math({self.name}.{func.__name__}(Dataset({other.name}))."
            )
            ts_logger.debug(
                f"DATASET {self.name}: .math({self.name},{other.name}) redirect to operating on {other.name}.data."
            )
            out_data = self.data.copy()
            out_data[self.numeric_columns()] = func(
                self.data[self.numeric_columns()], other.data[other.numeric_columns()]
            )
            other_name = other.name
            other_as_of = other.as_of_utc

        elif isinstance(other, pd.DataFrame):
            # element-wise matrix operation
            ts_logger.debug(
                f"DATASET {self.name}: .math({self.name}.{func.__name__}(pd.dataframe)."
            )

            out_data = self.data.copy()
            num_cols = self.numeric_columns()
            out_data[num_cols] = func(out_data[num_cols], other[num_cols])

            other_name = "df"
            other_as_of = None

        elif isinstance(other, int | float):
            out_data = self.data.copy()

            num_cols = self.numeric_columns()
            out_data[num_cols] = func(out_data[num_cols], other)

            other_name = str(other)
            other_as_of = None

        elif isinstance(other, np.ndarray):
            # TODO: this needs more thorough testing!
            # Compare shape of the ndarray against the numeric_columns of self.data. There are up to 3 accepted cases (depending on the operation):
            #  * matrix;         shape = (data.numeric.rows, data.numeric.columns)
            #  * column vector;  shape = (data.numeric.rows, 1)
            #  * row vector;     shape = (1, data.numeric.columns)
            #
            if other.ndim == 1 and (
                other.shape[0] == len(self.data)
                or other.shape[0] == len(self.data.columns)
            ):
                out_data = self.data.copy()

                out_data[self.numeric_columns()] = func(
                    out_data[self.numeric_columns()], other
                )
            else:
                raise ValueError(
                    f"Incompatible shapes for element-wise {func.__name__}"
                )
            other_name = "ndarray"
            other_as_of = None
        else:
            raise ValueError("Unsupported operand type")

        if other_as_of:
            out_as_of = max(self.as_of_utc, other_as_of)
        else:
            out_as_of = self.as_of_utc

        out = self.__class__(
            name=f"({self.name}.{func.__name__}.{other_name})",
            data_type=self.data_type,
            as_of_tz=out_as_of,
            data=out_data,
        )
        # TODO: Should also update series names and set/series tags.
        ts_logger.debug(
            f"DATASET.math({func.__name__}, {self.name}, {other_name}) --> {out.name}\n\t{out.data}."
        )
        return out

    def __add__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Add two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.add)

    def __radd__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right add two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.add)

    def __sub__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Subtract two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.subtract)

    def __rsub__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right subtract two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.subtract)

    def __mul__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Multiply two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.multiply)

    def __rmul__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right multiply two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.multiply)

    def __truediv__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Divide two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.divide)

    def __rtruediv__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right divide two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.divide)

    def __floordiv__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Floor divide two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.floor_divide)

    def __rfloordiv__(
        self, other: Self | pd.DataFrame | pd.Series | int | float
    ) -> Any:
        """Right floor divide two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.floor_divide)

    def __pow__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Power of two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.power)

    def __rpow__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right power of two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.power)

    def __mod__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Modulo of two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.mod)

    def __rmod__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Right modulo of two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.mod)

    @no_type_check
    def __eq__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Check equality of two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.equal)

    def __gt__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Check greater than for two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.greater)

    def __lt__(self, other: Self | pd.DataFrame | pd.Series | int | float) -> Any:
        """Check less than for two datasets or a dataset and a dataframe, numpy array or scalar."""
        return self.math(other, np.less)

    def __repr__(self) -> str:
        """Returns a machine readable string representation of Dataset, ideally sufficient to recreate object."""
        return f'Dataset(name="{self.name}", data_type={self.data_type!r}, as_of_tz="{self.as_of_utc.isoformat()}")'

    def __str__(self) -> str:
        """Returns a human readable string representation of the Dataset."""
        return str(
            {
                "name": self.name,
                "data_type": str(self.data_type),
                "as_of_utc": self.as_of_utc,
                "series": str(self.series),
                "data": self.data.size,
            }
        )

    def aggregate(
        self,
        attributes: list[str],
        taxonomies: list[int | meta.Taxonomy | dict[str, str] | PathStr],
        functions: list[str | F] | set[str | F],
        sep: str = "_",
    ) -> Self:
        """Aggregate dataset by taxonomy hierarchies.

        Args:
            attributes (list[str]): The attributes to aggregate by.
            taxonomies (list[int | meta.Taxonomy | dict[str, str] | PathStr]):
                Value definitions for `attributes` can be either ::py:class:`meta.Taxonomy` objects,
                or klass_ids, data dictionaries or paths that can be used to retrieve or construct them.
            functions (list[str|F] | set[str|F]): Optional function name (or list) of the function names to apply (mean | count | sum | ...). Defaults to `sum`.
            sep (str): Optional separator used when joining multiple attributes into names of aggregated series. Defaults to '_'.

        Returns:
            Self: A dataset object with the aggregated data.
            If the taxonomy object has hierarchical structure, aggregate series are calculated for parent nodes at all levels.
            If the taxonomy is a flat list, only a single `total` aggregate series is calculated.

        Raises:
            TypeError: If any of the taxonomy identifiere are of unexpected types.

        Examples:
            To calculate 10 and 90 percentiles and median for the dataset `x` where codes from KLASS 157 (energy_balance) distinguishes between series in the set.

            >>> from ssb_timeseries.dataset import Dataset
            >>> from ssb_timeseries.properties import SeriesType
            >>> from ssb_timeseries.sample_data import create_df
            >>> from ssb_timeseries.meta import Taxonomy
            >>>
            >>> klass157 = Taxonomy(klass_id=157)
            >>> klass157_leaves = [n.name for n in klass157.structure.root.leaves]
            >>> tag_permutation_space = {"A": klass157_leaves, "B": ["q"], "C": ["z"]}
            >>> series_names: list[list[str]] = [value for value in tag_permutation_space.values()]
            >>> sample_df = create_df(*series_names, start_date="2024-01-01", end_date="2024-12-31", freq="MS",)
            >>> sample_set = Dataset(name="sample_set",
            >>>     data_type=SeriesType.simple(),
            >>>     data=sample_df,
            >>>     name_pattern=["A", "B", "C"],
            >>> )
            >>>
            >>> def perc10(x):
            >>>     return x.quantile(.1, axis=1, numeric_only=True, interpolation="linear")
            >>>
            >>> def perc90(x):
            >>>     return x.quantile(.9, axis=1, numeric_only=True, interpolation="linear")
            >>>
            >>> percentiles = sample_set.aggregate(["energy_balance"], [157], [perc10, 'median', perc90])
        """
        taxonomy_dict = {}
        for name, t in zip(attributes, taxonomies, strict=False):
            if isinstance(t, meta.Taxonomy):
                obj = t
            elif isinstance(t, int):
                obj = meta.Taxonomy(klass_id=t)
            elif isinstance(t, dict):
                obj = meta.Taxonomy(data=t)
            elif isinstance(t, PathStr):
                obj = meta.Taxonomy(path=t)
            else:
                raise TypeError(
                    f"Taxonomy object or valid identifier expected, got {type(t)}"
                )
            taxonomy_dict[name] = obj

        df = self.data.copy().drop(columns=self.numeric_columns())
        for p in meta.permutations(taxonomy_dict, "parents"):
            criteria = {}
            for attr, value in p.items():
                criteria[attr] = taxonomy_dict[attr].leaf_nodes(value)
            # criteria = {{attr, taxonomy_dict[attr].leaf_nodes(value)} for attr, value in p.items()}
            ts_logger.debug(f"{criteria=}")
            output_series_name = sep.join(p.values())
            leaf_node_subset = self.filter(tags=criteria, output="df")
            for func in functions:
                if isinstance(func, str):
                    new_col_name = f"{func}({output_series_name})"
                elif isinstance(func, list):
                    new_col_name = f"{func[0]}{func[1]}({output_series_name})"  # type: ignore[unreachable]
                else:
                    new_col_name = f"{func.__name__}({output_series_name})"
                df[new_col_name] = column_aggregate(leaf_node_subset, func)

        return self.copy(f"{self.name}.{functions}", data=df)


def column_aggregate(df: pd.DataFrame, method: str | F) -> pd.Series | Any:
    """Helper function to calculate aggregate over dataframe columns."""
    # ts_logger.debug(f"DATASET.column_aggregate '{method}' over columns:\n{df.columns}")
    if isinstance(method, F):  # type: ignore
        out = method(df)  # type: ignore[operator]
    else:
        match method:
            case "mean" | "average":
                out = df.mean(axis=1, numeric_only=True)
            case "min" | "minimum":
                out = df.min(axis=1, numeric_only=True)
            case "max" | "maximum":
                out = df.max(axis=1, numeric_only=True)
            case "count":
                out = df.count(axis=1, numeric_only=True)
            case "sum":
                out = df.sum(axis=1, numeric_only=True)
            case "median":
                out = df.quantile(0.5, axis=1, numeric_only=True)
            case ["quantile", *params] | ["percentile", *params]:
                if len(params) > 1:  # type: ignore[unreachable]
                    interpolation: str = params[1]
                else:
                    interpolation = "linear"
                quantile = float(params[0])
                if quantile > 1:
                    quantile /= 100
                out = df.quantile(
                    quantile,
                    axis=1,
                    numeric_only=True,
                    interpolation=interpolation,
                )
            case _:
                raise NotImplementedError(
                    f"Aggregation method '{method}' is not implemented (yet)."
                )
    return out


def search(
    pattern: str = "*", as_of_tz: datetime = None
) -> list[io.SearchResult] | Dataset | list[None]:
    """Search for datasets by name matching pattern."""
    found = io.find_datasets(pattern=pattern)
    ts_logger.debug(f"DATASET.search returned:\n{found} ")

    if len(found) == 1:
        return Dataset(
            name=found[0].name,
            data_type=properties.seriestype_from_str(found[0].type_directory),
            as_of_tz=as_of_tz,
        )
    else:
        return found


def catalog_search(
    pattern: meta.TagDict,
    as_of_tz: datetime = None,
    object_type: str | list[str] = "dataset",
) -> list[io.SearchResult] | Dataset | list[None]:
    """Search across datasets by tags pattern."""
    found = io.find_datasets(pattern=pattern)
    ts_logger.debug(f"DATASET.search returned:\n{found} ")

    if len(found) == 1:
        return Dataset(
            name=found[0].name,
            data_type=properties.seriestype_from_str(found[0].type_directory),
            as_of_tz=as_of_tz,
        )
    else:
        return found



# ruff: noqa
# type: ignore #NOSONAR
"""
Helper module for date and time utility functions.

Notable examples include converting between UTC and local time, standardised string formats for file names, and (planned for future use) intervals.
"""

from datetime import datetime as dt, tzinfo
from typing import Any
from zoneinfo import ZoneInfo

from dateutil import parser

from ssb_timeseries.logging import ts_logger


MAX_TIME_PRECISION = "second"
DEFAULT_TIMESPEC = "seconds"
DEFAULT_TZ = ZoneInfo("Europe/Oslo")  # Will shift between CET and CEST
UTC = ZoneInfo("UTC")


def date_utc(some_date: dt | str | None, **kwargs) -> dt:
    """Convert date to UTC.

    If date has no timezone information, the data is assumed to be in CET.

    The output will be rounded to the precision specified by kwarg 'rounding'. Max precision 'second' will be used if none is provided.
    """
    if some_date is None or some_date == "":
        return date_round(now_utc())
    else:
        dt_type = ensure_datetime(some_date)
        assert isinstance(dt_type, dt)
        tz_aware = ensure_tz_aware(dt_type)
        correct_tz = tz_aware.astimezone(tz=UTC)
        return date_round(correct_tz, **kwargs)


def date_local(some_date: dt | str, **kwargs) -> dt:
    """Convert date to Europe/Oslo timezone; ie shifting between CET and CEST.

    If the date has no timezone information, the data is assumed to be in Oslo timezone.

    The output will be rounded to the precision specified by kwarg 'rounding'. Default precision 'minute' will be used if none is provided.
    """
    dt_type = ensure_datetime(some_date, tz=DEFAULT_TZ)
    return date_round(dt_type, **kwargs)


def date_round(d: dt, **kwargs) -> dt:
    """Round date to specified by kwarg 'rounding' or default precision.

    Rounding can take the values 'none', 'day', 'd', 'hour', 'h', 'minute', 'min', 'm', 'second', 'sec', or 's'.

    Default precision 'minute' is used if none is provided.
    """

    rounding = kwargs.get("rounding", MAX_TIME_PRECISION)
    match rounding.lower():
        case "day" | "d":
            out = d.replace(hour=0, minute=0, second=0, microsecond=0)
        case "hour" | "h":
            out = d.replace(minute=0, second=0, microsecond=0)
        case "minute" | "min" | "m":
            out = d.replace(second=0, microsecond=0)
        case "second" | "sec" | "s":
            out = d.replace(microsecond=0)
        case "none":
            out = d
    return out


def ensure_datetime(some_date_representation: Any, **kwargs) -> dt:
    """Make sure that we are dealing with a datetime object, convert if possible.

    If input is None or empty strings will be converted to now_utc().
    """
    if isinstance(some_date_representation, dt):
        return some_date_representation
    elif some_date_representation is None or some_date_representation == "":
        date_as_dt = now_utc(**kwargs)
    else:
        try:
            date_as_dt = some_date_representation.to_datetime()
        except (ValueError, TypeError, AttributeError):
            # except AttributeError:
            # d = dt.strptime(d, "%Y-%m-%d")
            date_as_dt = parser.parse(some_date_representation)
        return date_as_dt
    return date_as_dt


def ensure_tz_aware(some_date: dt) -> dt:
    """Make sure that our datetime object is timezone aware.

    Assume CET if timezone information is missing.
    """
    # if some_date.tzinfo is None or some_date.tzinfo.utcoffset(some_date) is None:
    if is_tz_naive(some_date):
        ts_logger.debug(
            "DATE_UTC catched a date without timezone info. This will become an error later. Assuming CET."
        )
        try:
            tz_aware = some_date.astimezone(tz=DEFAULT_TZ)
        except (ValueError, TypeError, AttributeError):
            tz_aware = some_date.replace(tzinfo=DEFAULT_TZ)
        return tz_aware
    else:
        return some_date


def is_tz_aware(d: dt) -> bool:
    return d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None


def is_tz_naive(d: dt) -> bool:
    return d.tzinfo is None or d.tzinfo.utcoffset(d) is None


def now_utc(**kwargs) -> dt:
    """Return now in UTC."""
    t = dt.now(tz=UTC)
    return date_round(t, **kwargs)


def now_cet(**kwargs) -> dt:
    """Return now in CET."""
    t = dt.now(tz=UTC)
    return date_round(t, **kwargs)


def utc_iso(d: Any, timespec: str = DEFAULT_TIMESPEC) -> str:
    """Convert date to UTC and return as an ISO formatted string."""
    return date_utc(d).isoformat(timespec=timespec)


def utc_iso_no_colon(d: dt, timespec: str = DEFAULT_TIMESPEC) -> str:
    """Convert date to UTC and return as an ISO formatted string without the colons."""
    return utc_iso(d, timespec=timespec).replace(":", "")


class Interval:
    def __init__(self, f=None, t=None, **kwargs) -> None:
        """Interval(date_from, date_to)
        or a number of variations of named parameters: start/stop, begin/end, as_of, as_of_from/as_of_to, valid_from, valid_to -
        If only "as_of" is provided, start/stop are both set to this date.
        Interval.start defaults to datetime.min
        Interval.stop defaults to datetime.max
        """
        as_of = kwargs.get("as_of")
        as_of_from = kwargs.get("as_of_from")
        as_of_to = kwargs.get("as_of_to")
        valid_from = kwargs.get("valid_from")
        valid_to = kwargs.get("valid_to")
        start = kwargs.get("start")
        stop = kwargs.get("stop")
        from_ = kwargs.get("from_date")
        to = kwargs.get("to_date")
        date_from = kwargs.get("date_from")
        date_to = kwargs.get("date_to")
        from_date = kwargs.get("from_date")
        to_date = kwargs.get("to_date")
        begin = kwargs.get("begin")
        end = kwargs.get("end")

        precision: str = kwargs.get("precision", MAX_TIME_PRECISION)

        ts_logger.debug(f"Interval.__init__ with kwargs:\n{kwargs}")

        self.start = (
            f
            or as_of
            or start
            or as_of_from
            or from_
            or date_from
            or from_date
            or valid_from
            or begin
            or dt.min
        )
        self.stop = (
            t
            or as_of
            or stop
            or to
            or as_of_to
            or date_to
            or to_date
            or valid_to
            or end
            or dt.max
        )

        if precision:
            # "round" to MAX_TIME_precision if provided
            ...

        ts_logger.debug(f"Interval.__init__ returns self:\n{self}")

    def includes(self, *args: dt | list[dt]):
        ts_logger.debug(f"Interval.include args: {args}")
        if len(args) == 1:
            ts_logger.debug("Interval.include - single input")
            out: bool = args[0] >= self.start and args[0] <= self.stop
        else:
            out = [x >= self.start and x <= self.stop for x in args]
        ts_logger.debug(f"Interval.include returns:\n{out}")
        return out

    def all(self):
        self.start = dt.min
        self.stop = dt.max

    def __eq__(self, other) -> str:
        return (self.start, self.stop) == (other.start, other.stop)

    def __lt__(self, other) -> str:
        ts_logger.debug(
            f"Interval.lt: \n\t{self.stop} < ({other.start} < {other.stop})"
        )
        return self.stop < other.start

    def __gt__(self, other) -> str:
        ts_logger.debug(f"Interval.gt: \n\t{self.start} > {other.stop}")
        return self.start > other.stop

    def __le__(self, other) -> str:
        ts_logger.debug(
            f"Interval.le: \n\t{self.start} - {self.stop}\n\t{other.start} - {other.stop}"
        )
        return (self.start, self.stop) <= (other.start, other.stop)

    def __ge__(self, other) -> str:
        ts_logger.debug(
            f"Interval.ge: \n\t{self.start} - {self.stop}\n\t{other.start} - {other.stop}"
        )
        return (self.start, self.stop) >= (other.start, other.stop)

    def __repr__(self) -> str:
        expr = f"Interval(from_date = {self.start}, to_date = {self.stop}) "
        return expr

    def __str__(self) -> str:
        expr = f"Interval from {self.start} to {self.stop}."
        return expr



"""Configurations for the SSB timeseries library.

An environment variable TIMESERIES_CONFIG is expected to point to a JSON file with configurations.
If these exist, they will be loaded and put into a Config object CONFIG when the configuration module is loaded.

In most cases, this would happen behind the scene when :py:mod:`ssb_timeseries.dataset` or :py:mod:`ssb_timeseries.catalog` are imported.

Directly accessing the configuration module should only be required when manipulating configurations from Python code.

Example:
    >>> # doctest: +SKIP
    >>> from ssb_timeseries.config import CONFIG
    >>> CONFIG.catalog = 'gs://{bucket}/timeseries/metadata/'
    >>> CONFIG.save()
    >>> # doctest: -SKIP

For switching between preset configurations, use the `timeseries-config` command::

    poetry run timeseries-config <option>

which is equivalent to::

    python ./config.py <option>

See :py:func:`ssb_timeseries.config.main` for details on the named options.

"""

import json
import logging
import os
import sys
from pathlib import Path

from jsonschema import validate
from jsonschema.exceptions import ValidationError as JsonValidationError
from typing_extensions import Self

from ssb_timeseries import fs
from ssb_timeseries.types import PathStr

# mypy: disable-error-code="assignment, arg-type, override,call-arg,has-type,no-untyped-def,attr-defined,import-untyped,"

PACKAGE_NAME = "ssb_timeseries"
ENV_VAR_NAME = "TIMESERIES_CONFIG"


def unset_env_var() -> str:
    """Unsets the environment variable :py:const:`ENV_VAR_NAME` and returns the value that was unset."""
    return os.environ.pop(ENV_VAR_NAME, "")


def active_file(path: PathStr = "") -> str:
    """If a path is provided, sets environment variable :py:const:`ENV_VAR_NAME` to specify the location of the configuration file.

    Returns the value of the environment variable by way of :py:func:`get_active_file`.
    """
    if path:
        os.environ[ENV_VAR_NAME] = str(path)
        logging.debug(f"Set environment variable {ENV_VAR_NAME} to {path}")

    return os.environ.get(ENV_VAR_NAME, "")


# failed to make tests fail on warning???
# logging.warning("want this to make tests fail - but they do not")
# warnings.warn('this does!')

HOME = str(Path.home())
SHARED_PROD = "gs://ssb-prod-dapla-felles-data-delt/tidsserier"
SHARED_TEST = "gs://ssb-test-dapla-felles-data-delt/tidsserier"
GCS = SHARED_PROD

JOVYAN = "/home/jovyan"
DAPLALAB_HOME = "/home/onyxia/work"
SSB_DIR_NAME = "tidsserier"
ROOT_DIR_NAME = "tidsserier"
META_DIR_NAME = "metadata"
SSB_CONF_DIR = "konfigurasjon"
LINUX_CONF_DIR = ".config"
SSB_LOGDIR = "logger"
LOGDIR = "logs"
LOGFILE = "timeseries.log"
CONFIGFILE = "timeseries_config.json"

DAPLA_TEAM_CONTEXT = os.getenv("DAPLA_TEAM_CONTEXT", "")
DAPLA_ENV = os.getenv("DAPLA_ENVIRONMENT", "")
"""Returns the Dapla environment: 'prod' | test | dev"""
DAPLA_TEAM = os.getenv("DAPLA_TEAM", "")
"""Returns the Dapla team/project name.'"""
DAPLA_BUCKET = f"gs://{DAPLA_TEAM}-{DAPLA_ENV}"
"""Returns the Dapla product bucket name for the current environment: gs://{DAPLA_TEAM}-{DAPLA_ENV}."""
PRESETS: dict[str, dict] = {
    "default": {
        "configuration_file": str(Path(HOME, LINUX_CONF_DIR, PACKAGE_NAME, CONFIGFILE)),
        "bucket": str(Path(HOME)),
        "timeseries_root": str(Path(HOME, ROOT_DIR_NAME)),
        "catalog": str(Path(HOME, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(HOME, LOGDIR, LOGFILE)),
    },
    "home": {
        "configuration_file": str(Path(HOME, LINUX_CONF_DIR, PACKAGE_NAME, CONFIGFILE)),
        "bucket": str(Path(HOME)),
        "timeseries_root": str(Path(HOME, ROOT_DIR_NAME)),
        "catalog": str(Path(HOME, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(HOME, ROOT_DIR_NAME, LOGDIR, LOGFILE)),
    },
    "gcs": {
        "configuration_file": str(Path(GCS, ROOT_DIR_NAME, CONFIGFILE)),
        "bucket": str(Path(GCS)),
        "timeseries_root": str(Path(GCS, ROOT_DIR_NAME)),
        "catalog": str(Path(HOME, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(HOME, ROOT_DIR_NAME, LOGFILE)),
    },
    "shared-test": {
        "configuration_file": str(Path(HOME, SSB_CONF_DIR, PACKAGE_NAME, CONFIGFILE)),
        "bucket": str(Path(SHARED_TEST)),
        "timeseries_root": str(Path(SHARED_TEST, ROOT_DIR_NAME)),
        "catalog": str(Path(SHARED_TEST, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(SHARED_TEST, LOGDIR, LOGFILE)),
    },
    "shared-prod": {
        "configuration_file": str(
            Path(SHARED_PROD, SSB_CONF_DIR, PACKAGE_NAME, CONFIGFILE)
        ),
        "bucket": str(Path(SHARED_PROD)),
        "timeseries_root": str(Path(SHARED_PROD, ROOT_DIR_NAME)),
        "catalog": str(Path(SHARED_PROD, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(SHARED_PROD, LOGDIR, LOGFILE)),
    },
    "jovyan": {
        "configuration_file": str(
            Path(JOVYAN, LINUX_CONF_DIR, PACKAGE_NAME, CONFIGFILE)
        ),
        "bucket": str(Path(JOVYAN)),
        "timeseries_root": str(Path(JOVYAN, ROOT_DIR_NAME)),
        "catalog": str(Path(HOME, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(JOVYAN, LOGDIR, LOGFILE)),
    },
    "dapla": {
        "configuration_file": str(
            Path(DAPLA_BUCKET, SSB_CONF_DIR, PACKAGE_NAME, CONFIGFILE)
        ),
        "bucket": str(Path(JOVYAN)),
        "timeseries_root": str(Path(DAPLA_BUCKET, ROOT_DIR_NAME)),
        "catalog": str(Path(DAPLA_BUCKET, ROOT_DIR_NAME, META_DIR_NAME)),
        "log_file": str(Path(DAPLA_BUCKET, SSB_LOGDIR, LOGFILE)),
    },
}

PRESETS["defaults"] = PRESETS["default"]
DEFAULTS = PRESETS["default"]


class Config:
    """Configuration class; for reading and writing timeseries configurations.

    If instantiated with no parameters, an existing configuration file is exepected to exist: either in a location specified by the environment variable TIMESERIES_CONFIG or in the default location in the user's home directory. If not, an error is returned.

    If the :py:attr:`configuration_file` attribute is specified, configurations will be loaded from that file. No other parameters are required. A :py:exc:`FileNotFoundError` or :py:exc:`FileDoesNotExist` error will be returned if the file is not found. In this case, no attempt is made to load configurations from locations specified by environment variable or defaults.

    If any additional parameters are provided, they will override values from the configuration file. If the result is not a valid configuration, a ValidationError is raised.

    If one or more parameters are provided, but the `configuration_file` parameter is not among them, configurations are identified by the environment variable TIMESERIES_CONFIG or the default configuration file location (in that order of priority). Provided parameters override values from the configuration file. If the result is not a valid configuration, an error is raised.

    The returned configuration will not be saved, but held in memory only till the :py:meth:`save` method is called. Then the configuration will be savedto a file and the environment variable TIMESERIES_CONFIG set to reflect the location of the file.

    """

    configuration_file: PathStr
    """The path to the configuRation file."""
    timeseries_root: PathStr
    """The root directory for data storage of a repository."""
    catalog: PathStr
    """The path to the metadata directory of a repository ."""
    log_file: PathStr
    """The path to the log file."""
    bucket: PathStr
    """The topmost level of the GCS bucket for the team."""

    def __init__(self, **kwargs) -> None:  # noqa: D417, ANN003, DAR101, DAR402, RUF100
        """Initialize Config object from keyword arguments.

        Keyword Arguments:
            preset (str): Optional. Name of a preset configuration. If provided, the preset configuration is loaded, and no other parameters are considered.
            configuration_file (str): Path to the configuration file. If the parameter is not provided, the environment variable TIMESERIES_CONFIG is used. If the environment variable is not set, the default configuration file location is used.
            timeseries_root (str): Path to the root directory for time series data. If one of these identifies a vaild json file, the configuration is loaded from that file and no other parameters are required. If provided, they will override values from the configuration file.
            catalog (str): Path to the catalog file.
            log_file (str): Path to the log file.
            bucket (str): Name of the GCS bucket.

        Raises:
            :py:exc:`FileNotFoundError`: If the configuration file as implied by provided or not provided parameters does not exist.   # noqa: DAR402
            :py:exc:`ValidationError`: If the resulting configuration is not valid.   # noqa: DAR402
            :py:exc:`EnvVarNotDefinedeError`: If the environment variable TIMESERIES_CONFIG is not defined.

        Examples:
            Load an existing config from TIMESERIES_CONFIG or default location:

                >>> from ssb_timeseries.config import Config
                >>> config = Config.active()
        """
        preset_name = kwargs.pop("preset", "")
        ignore_file = kwargs.pop("ignore_file", False)
        param_specified_config_file = kwargs.get("configuration_file", "")
        kwargs_are_complete_config = is_valid_config(kwargs)[0]

        if preset_name:
            logging.debug(f"Loading preset configuration {preset_name}.")
            self.apply(PRESETS[preset_name])
            return
        elif kwargs_are_complete_config:
            logging.debug("Complete configuration in parameters.")
            self.apply(kwargs)
            return
        elif param_specified_config_file:
            logging.debug(f"Loading configuration from {param_specified_config_file}")
            if set(kwargs.keys()) == set(["configuration_file"]):
                no_file_is_an_error = True
            else:
                no_file_is_an_error = not is_valid_config(kwargs)[0]

            if not ignore_file:
                config_from_file = load_json_file(
                    path=param_specified_config_file,
                    error_on_missing=no_file_is_an_error,
                )
            else:
                config_from_file = {}

            config_values = PRESETS["default"]
            config_values.update(config_from_file)
            logging.debug(f"{config_values=}")
        elif active_file():
            # if the path is specified by the environment variable, not finding it is an error
            logging.debug(f"Loading configuration from {active_file()}")
            config_values = load_json_file(
                path=active_file(),
                error_on_missing=True,
            )
        # elif not active_file():
        #    raise MissingEnvironmentVariableError
        else:
            logging.warning(
                f"The environment variable {ENV_VAR_NAME} did not exist and no configuration file parameter was provided. Loading default configuration."
            )
            config_values = PRESETS["defaults"]

        config_values.update(kwargs)

        self.apply(config_values)

    def apply(self, configuration: dict) -> None:
        """Set configuration values from a dictionary."""
        is_valid, reason = is_valid_config(configuration=configuration)
        if is_valid:
            for key, value in configuration.items():
                setattr(self, key, value)
        else:
            logging.error(f"Invalid configuration {configuration}\n{reason}.")
            raise ValidationError(f"Invalid configuration {configuration}.")

    @property
    def is_valid(self) -> bool:
        """Check if the configuration has all required fields."""
        result: bool = is_valid_config(self.__dict__())[0]
        return result

    def save(self, path: PathStr = "") -> None:
        """Saves configurations to the JSON file defined by `path` or :py:attr:`configuration_file`.

        If `path` is set, it will take presence and :attr:`.configuration_file` will be set accordingly.

        Args:
            path (PathStr): Full path of the JSON file to save to. If not specified, it will attempt to use the environment variable TIMESERIES_CONFIG before falling back to the default location `$HOME/.config/ssb_timeseries/timeseries_config.json`.

        Raises:
            ValueError: If `path` is not provided and :attr:`configuration_file` is not set.
        """
        if path:
            self.configuration_file = str(path)
        elif not self.configuration_file:
            raise ValueError(
                "Configuration file must have a value or path must be specified."
            )
        else:
            path = self.configuration_file

        fs.write_text(content=str(self), path=str(path), file_format="json")
        if not fs.exists(self.log_file):
            fs.touch(self.log_file)

        active_file(str(path))

    def __getitem__(self, item: str) -> str:
        """Get the value of a configuration."""
        return str(getattr(self, item))

    def __eq__(self, other: Self | dict) -> bool:
        """Equality test."""
        if isinstance(other, dict):
            return self.__dict__() == other
        else:
            return self.__dict__() == other.__dict__()

    def __dict__(self) -> dict:
        """Return timeseries configurations as JSON string."""
        fields = [
            "configuration_file",
            "timeseries_root",
            "catalog",
            "log_file",
            "bucket",
        ]
        out = {}
        for field in fields:
            out[field] = self[field]
        return out

    def __str__(self) -> str:
        """Return timeseries configurations as JSON string."""
        return json.dumps(self.__dict__(), sort_keys=True, indent=2)

    @classmethod
    def active(cls) -> Self:
        """Force reload and return the configuration identified by :py:const:`ENV_VAR_NAME`."""
        return cls(configuration_file=active_file())


class MissingEnvironmentVariableError(Exception):
    """The environment variable TIMESEREIS_CONFIG must be defined."""

    ...


class ValidationError(Exception):
    """Configuration validation error."""

    ...


def load_json_file(path: PathStr, error_on_missing: bool = False) -> dict:
    """Read configurations from a JSON file into a Config object."""
    if fs.exists(path):
        from_json = fs.read_json(path)
        if not isinstance(from_json, dict):
            return json.loads(from_json)  # type:ignore
        else:
            return from_json

    elif error_on_missing:
        raise FileNotFoundError(
            f"A configuration file at {path} file was specified, but does not exist."
        )
    else:
        return {}


def migrate_to_new_config_location(file_to_copy: PathStr = "") -> str:
    """Copy existing configuration files to the new default location $HOME/.config/ssb_timeseries/.

    The first file copied will be set to active.

    Args:
        file_to_copy (PathStr): Optional. Path to a existing configuration file. If not provided, the function will look in the most common location for SSBs old JupyterLab and DaplaLab.
    """
    DEFAULTS["configuration_file"]

    if file_to_copy:
        fs.cp(file_to_copy, DEFAULTS["configuration_file"])
        return str(file_to_copy)
    else:
        copy_these = [
            {
                "replace": "_active",
                "source": active_file(),
            },
            {
                "replace": "_home",
                "source": path_str(HOME, "timeseries_config.json"),
            },
            {
                "replace": "_jovyan",
                "source": "/home/joyan/timeseries_config.json",
            },
            {
                "replace": "_daplalab",
                "source": "/home/joyan/work/timeseries_config.json",
            },
        ]
        copied = []
        not_found = []

        for c in copy_these:
            if fs.exists(c["source"]):
                # copy all to .config, but let filename signal where it was copied from
                target = DEFAULTS["configuration_file"].replace(
                    ".json", f"{c['replace']}.json"
                )
                fs.cp(c["source"], target)
                copied.append(target)
            else:
                not_found.append(c["source"])
        else:
            logging.warning(f"Configuration files were not found: {not_found}.")

    if copied:
        # copy the first file = make it the active one
        fs.cp(copied[0], DEFAULTS["configuration_file"])
        active_file(copied[0])
        logging.info(
            f"Configuration files were copied: {copied}.\nActive: {copied[0]}."
        )
        return str(copied[0])
    else:
        # no files were found --> create one from defaults?
        # new = Config(preset="default")
        # ew.save()
        return ""


def configuration_schema(version: str = "0.3.1") -> dict:
    """Return the JSON schema for the configuration file."""
    match version:
        case "0.3.1" | "0.3.2" | _:
            cfg_schema = {
                "$schema": "https://json-schema.org/draft/2020-12/schema",
                "type": "object",
                "properties": {
                    "bucket": {"type": "string"},
                    "configuration_file": {"type": "string"},
                    "log_file": {"type": "string"},
                    "catalog": {"type": "string"},
                    "timeseries_root": {"type": "string"},
                },
                "required": [
                    "bucket",
                    "configuration_file",
                    "log_file",
                    "catalog",
                    "timeseries_root",
                ],
            }
    logging.debug(f"Schema {version}:{cfg_schema}")
    return cfg_schema


class DictObject(object):  # noqa
    """Helper class to convert dict to object."""

    def __init__(self, dict_: dict) -> None:  # noqa: D107
        self.__dict__.update(dict_)

    @classmethod
    def from_dict(cls, d: dict):  # noqa: ANN206, D102
        return json.loads(json.dumps(d), object_hook=DictObject)


def is_valid_config(configuration: dict) -> tuple[bool, object]:
    """Check if a dictionary is a valid configuration.

    A valid configuration has the same keys as DEFAULTS.
    """
    try:
        out = (True, None)
        validate(
            instance=configuration,
            schema=configuration_schema(),
        )
    except JsonValidationError as err:
        logging.debug(f"Invalid configuration {configuration}\nValidation error. {err}")
        out = (False, err)
    return out


def presets(named_config: str) -> dict:  # noqa: RUF100, DAR201
    """Set configurations to predefined defaults.

    Raises:
        ValueError: If args is not 'home' | 'gcs' | 'jovyan'.
    """
    if named_config in PRESETS:
        return PRESETS[named_config]
    else:
        raise ValueError(
            f"Named configuration preset '{named_config}' was not recognized."
        )


def main(*args: str | PathStr) -> None:
    """Set configurations to predefined defaults when run from command line.

    Use:
        ```
        poetry run timeseries-config <option>
        ```
    or
        ```
        python ./config.py <option>`
        ```

    Args:
        *args (str): 'home' | 'gcs' | 'jovyan'.

    Raises:
        ValueError: If args is not 'home' | 'gcs' | 'jovyan'. # noqa: DAR402

    """
    if args:
        config_identifier: PathStr = args[0]
    else:
        config_identifier = sys.argv[1]

    print(f"Update configuration with named preset: '{config_identifier}'.")
    cfg = Config(preset=config_identifier)
    cfg.save(path=cfg.configuration_file)

    print(
        f"Preset configuration '{config_identifier}' was applied:\n\t{cfg.__dict__}\nSaved to file: {cfg.configuration_file}.\nEnvironment variable set: {os.getenv('TIMESERIES_CONFIG')=}"
    )


def path_str(*args) -> str:  # noqa: ANN002
    """Concatenate paths as string: str(Path(...))."""
    return str(Path(*args))


if __name__ == "__main__":
    """Execute when called directly, ie not via import statements."""
    # ??? `poetry run timeseries-config <option>` does not appear to go this route.
    # --> then it is not obvious that this is a good idea.
    print(f"Name of the script      : {sys.argv[0]=}")
    print(f"Arguments of the script : {sys.argv[1:]=}")
    main(sys.argv[1])
else:
    if active_file():
        if fs.exists(active_file()):
            CONFIGURATION_FILE = active_file()
        elif DAPLA_TEAM_CONTEXT:
            raise MissingEnvironmentVariableError(
                f"Environment variable {ENV_VAR_NAME} must be defined and point to a configuration file."
            )
        else:
            print(
                f"No configuration file was foumd at {active_file()}.\nOther locatsions will be tried. Files found will be copied to the default location and the first candidate will be set to active, ie copied once more to {DEFAULTS['configuration_file']}"
            )
            CONFIGURATION_FILE = migrate_to_new_config_location()
            if not fs.exists(CONFIGURATION_FILE):
                raise FileNotFoundError(
                    f"No configuration file was found at {active_file()}."
                )
    else:
        CONFIGURATION_FILE = ""  # PRESETS["defaults"]["configuration_file"]
        # str(DEFAULTS["configuration_file"])

    active_file(CONFIGURATION_FILE)
    CONFIG = Config(configuration_file=CONFIGURATION_FILE)
    """A Config object."""
    fs.touch(CONFIG.log_file)
    # do not save
    # CONFIG.save()



from collections.abc import Callable
from os import PathLike
from typing import TypeAlias

# from ssb_timeseries.logging import ts_logger

PathStr: TypeAlias = str | PathLike[str]
F: TypeAlias = Callable

# ruff: noqa: D202
